// DCC Headlight Decoder for DS-DCC decode
// By yaasan
// Based on Nicolas's sketch http://blog.nicolas.cx
// Inspired by Geoff Bunza and his 17 Function DCC Decoder & updated library


#include "NmraDcc_v141.h"
#include <avr/eeprom.h>	 //required by notifyCVRead() function if enabled below
#include <Servo.h>

// This Example shows how to use the library as a DCC Accessory Decoder or a DCC Signalling Decoder
// It responds to both the normal DCC Turnout Control packets and the newer DCC Signal Aspect packets 
// You can also print every DCC packet by uncommenting the "#define NOTIFY_DCC_MSG" line below
//#define DEBUG

#define MAX_SERVO 2
#define PIN_SERVO1 5
#define PIN_SERVO2 4

#define PIN_RELAY2   8  //R2基板以降対応
#define PIN_RELAY  A0 //R2基板以降対応

#define PIN_LED_DIV 6
#define PIN_LED_STR 7

#define PIN_POS1   A7 //Str, R2基板ではA7
#define PIN_POS2   A1 //Div, R2基板ではA1
#define PIN_SPD    A2 //R2基板ではA2

#define PIN_DCIN   2

#define PI_ADRBIT_0 1
#define PI_ADRBIT_1 0
#define PI_ADRBIT_2 A5
#define PI_ADRBIT_3 A4
#define PI_ADRBIT_4 A3
#define PI_ADRBIT_5 A6
#define PI_ADRBIT_6 9
#define PI_ADRBIT_7 10

#define ANA_ZERO 100

#define ANGLE_MIN 500 //1000 -90deg
#define ANGLE_MAX 2400 //2000 +90deg

struct SRV_PARAM
{
	int mOn;
	int mStatus;
	int mTargetPos;
	int current_position;
	int increment;
	int stop_value;
	int start_value;
};
SRV_PARAM *gServoParam = new SRV_PARAM[MAX_SERVO];

NmraDcc  Dcc ;
DCC_MSG  Packet ;
Servo myservo;  // create servo object to control a servo
Servo myservo2;  // create servo object to control a servo

struct CVPair
{
  uint16_t  CV;
  uint8_t   Value;
};
  
CVPair FactoryDefaultCVs [] =
{
  {CV_ACCESSORY_DECODER_ADDRESS_LSB, 1},
  {CV_ACCESSORY_DECODER_ADDRESS_MSB, 0},
  {CV_29_CONFIG, 0b10000000},
  {31, 10},    //  Increment
  {32, 0},   //  Start Position F0=0
  {33, 90},  //  End Position   F0=1
  {34, 0},   //  Current Position
  {35, 0},   // State
  {36, 10},    //  Increment
  {37, 0},   //  Start Position F0=0
  {38, 90},  //  End Position   F0=1
  {39, 0},   //  Current Position
	{40, 0}, // State
	{47, 0},    //  Address  1: 軸1と軸2は同期運転(アドレス同一)、0:軸1と軸2は個別運転(アドレスは連番)
	{52, 0},   //  1stアドレスオフセット
	{53, 0},   //  2ndアドレス
	{54, 0},   //  2ndアドレスオフセット

};

uint8_t FactoryDefaultCVIndex = 0;
unsigned long gPreviousWorkTime = 0;
uint16_t gAccessoryAddress = 1;
uint16_t gAccessoryAddress2 = 0;

//Task Schedule
unsigned long gPreviousL5 = 0;
unsigned long gPreviousL6 = 0;

uint8_t gCV47_AddrType = 0;
uint16_t gCV52_AddrOffset = 0;
uint16_t gCV53_Addr2 = 0;
uint16_t gCV54_AddrOffset2 = 0;


uint16_t getMyAddr_Acc(void);
void SetRelay(uint8_t inState);
void SetRelay2(uint8_t inState);
void SetLED(uint8_t inState);
uint16_t getMyAddrFromSW();


/* Functions  */



void notifyCVResetFactoryDefault()
{
  // Make FactoryDefaultCVIndex non-zero and equal to num CV's to be reset 
  // to flag to the loop() function that a reset to Factory Defaults needs to be done
  FactoryDefaultCVIndex = sizeof(FactoryDefaultCVs)/sizeof(CVPair);
};

const int DccAckPin = PIN_RELAY ;

// This function is called by the NmraDcc library when a DCC ACK needs to be sent
// Calling this function should cause an increased 60ma current drain on the power supply for 6ms to ACK a CV Read 
void notifyCVAck(void)
{
  // Serial.println("notifyCVAck") ;
  
  digitalWrite( PIN_RELAY, HIGH );
  digitalWrite( PIN_RELAY2, HIGH );
  delay( 10 );  
  digitalWrite( PIN_RELAY, LOW );
  digitalWrite( PIN_RELAY2, LOW );
}

// This function is called whenever a normal DCC Turnout Packet is received
void notifyDccAccState( uint16_t Addr, uint16_t BoardAddr, uint8_t OutputAddr, uint8_t State)
{
	uint16_t aAccAdress = Addr;
	uint8_t aDir = (OutputAddr & 0b001);
		
		//DEBUG
		// Serial.print("notifyDccAccState: ") ;
		// Serial.print(Addr,DEC) ;
		// Serial.print(",BAdr:");
		// Serial.print(BoardAddr,DEC) ;
		// Serial.print(",OAdr:");
		// Serial.print(OutputAddr,DEC) ;
		// Serial.print(", Target Adr:");
		// Serial.print(aAccAdress,DEC) ;
		// Serial.print(",");
		// Serial.print(State, HEX) ;
		// Serial.print("==");
		// Serial.print(gAccessoryAddress,DEC) ;
		// Serial.println(",");
	
	//アドレスチェック(CVから得た11bit分の信号を比較)
	if( gAccessoryAddress == aAccAdress)
	{
		
		if(aDir  == 0)
		{
			// Serial.println("Diverse(Axis1)") ;
		}
		else if( aDir == 1)
		{
			// Serial.println("Straight(Axis1)") ;
		}
		
		exec_function( 3,  aDir );
		
		/* ２軸目も同時に動かす */
		if( gCV47_AddrType == 1)
		{
			exec_function( 4,  aDir );
		}		
	}
	else if( ((gAccessoryAddress2) == aAccAdress) && (gCV47_AddrType == 0) )
	{
		if(aDir  == 0)
		{
			// Serial.println("Diverse(Axis2)") ;
		}
		else if( aDir == 1)
		{
			// Serial.println("Straight(Axis2)") ;
		}

		exec_function( 4,  aDir );
	}
	else
	{
		return;
	}
	
	


}

uint16_t getMyAddr_Acc(void)
{
  uint16_t  Addr ;
  uint8_t   CV29Value ;

  CV29Value = Dcc.getCV( CV_29_CONFIG ) ;

  if( CV29Value & 0b10000000 )  // Accessory Decoder? 
    Addr = ( Dcc.getCV( CV_ACCESSORY_DECODER_ADDRESS_MSB ) << 8 ) | Dcc.getCV( CV_ACCESSORY_DECODER_ADDRESS_LSB ) ;

  else   // Multi-Function Decoder?
  {
    if( CV29Value & 0b00100000 )  // Two Byte Address?
      Addr = ( Dcc.getCV( CV_MULTIFUNCTION_EXTENDED_ADDRESS_MSB ) << 8 ) | Dcc.getCV( CV_MULTIFUNCTION_EXTENDED_ADDRESS_LSB ) ;

    else
      Addr = Dcc.getCV( 1 ) ;
  }

  return Addr ;
}

// This function is called whenever a DCC Signal Aspect Packet is received
void notifyDccSigState( uint16_t Addr, uint8_t OutputIndex, uint8_t State)
{
  // Serial.print("notifyDccSigState: ") ;
  // Serial.print(Addr,DEC) ;
  // Serial.print(',');
  // Serial.print(OutputIndex,DEC) ;
  // Serial.print(',');
  // Serial.println(State, HEX) ;
}

void setup()
{
	//アクセサリアドレス(下位2bitを考慮)の先修得
	gAccessoryAddress = getMyAddrFromSW();
	
	
	if ( Dcc.getCV(CV_MULTIFUNCTION_PRIMARY_ADDRESS) == 0xFF )
	{
		//if eeprom has 0xFF then assume it needs to be programmed
		//Serial.println("CV Defaulting due to blank eeprom");
		notifyCVResetFactoryDefault();
	}
	else
	{
		//Serial.println("CV Not Defaulting");
	}
	
	// Serial.begin(9600);
	
#ifdef DEBUG
	// Serial.println("Turnout decoder(Servo)");
#endif
	
	// Setup which External Interrupt, the Pin it's associated with that we're using and enable the Pull-Up 
	Dcc.pin(0, 2, 0);

	// Call the main DCC Init function to enable the DCC Receiver
	Dcc.init( 140, 10, FLAGS_OUTPUT_ADDRESS_MODE | FLAGS_DCC_ACCESSORY_DECODER, 0 );
	
	
	/* アドレス読み出し */
	gCV47_AddrType  = Dcc.getCV( 47);
	gCV52_AddrOffset = Dcc.getCV( 52);
	
	if( gCV52_AddrOffset > 7)
	{
		gCV52_AddrOffset = 0;
	}
	
	gAccessoryAddress = getMyAddrFromSW() + (gCV52_AddrOffset << 8);
	
	/* 2nd Address */
	gCV53_Addr2 = Dcc.getCV( 53);
	gCV54_AddrOffset2 = Dcc.getCV( 54);
	
	if( gCV54_AddrOffset2 > 7)
	{
		gCV54_AddrOffset2 = 0;
	}
	
	if( (gCV53_Addr2 == 0) && (gCV54_AddrOffset2 == 0))
	{
		/* CV53=CV54=0のときは連番アドレスを有効にする。どちらかに値が入っている場合はCh2アドレスが有効になる。 */
		gAccessoryAddress2 = gAccessoryAddress + 1;
	}
	else
	{
		gAccessoryAddress2 = gCV53_Addr2 + (gCV54_AddrOffset2 << 8);
		
		if( gAccessoryAddress2 > 2044)
		{
			/* アドレスを連番に強制変更 */
			gAccessoryAddress2 = gAccessoryAddress + 1;
		}
		
	}
	
	
	
#ifdef DEBUG
	//アドレス表示
	// Serial.print("Acc Adr is ");
	// Serial.println(gAccessoryAddress);  
#endif


	for( int i = 0; i < 2; i++)
	{
		gServoParam[i].mOn = 0; //go to initial position
		gServoParam[i].mStatus = int (Dcc.getCV( 35 +(i * 5)));
		gServoParam[i].current_position = map(Dcc.getCV( 34 + (i * 5)), 0, 180, ANGLE_MIN,  ANGLE_MAX);
		gServoParam[i].stop_value = Dcc.getCV( 33 + (i * 5));
		gServoParam[i].start_value = Dcc.getCV( 32 + (i * 5));
		gServoParam[i].increment = int (Dcc.getCV( 31 + (i * 5)));
		
		if( gServoParam[i].mStatus > 1)
		{
			gServoParam[i].mStatus = 0;
		}
		
		if( gServoParam[i].stop_value > 180)
		{
			gServoParam[i].stop_value = 180;
		}
		
		if( gServoParam[i].start_value > 180)
		{
			gServoParam[i].start_value = 180;
		}
		
		
		
#ifdef DEBUG
		// Serial.print("Start Pos is ");
		// Serial.println(gServoParam[i].start_value);  
		// Serial.print("Stop Pos is ");
		// Serial.println(gServoParam[i].stop_value);  
#endif
	}
	
	//設定位置を上書き（手動用）
	gServoParam[0].stop_value = map( analogRead(PIN_POS2), 0, 1023, 0, 180 );
	gServoParam[0].start_value = map( analogRead(PIN_POS1), 0, 1023, 0, 180 );
	gServoParam[0].increment = GetIncremental();
	gServoParam[0].mOn = 0; //go to initial position
	
	
	//初期位置へ
	switch( gServoParam[0].mStatus)
	{
	case 0:
		gServoParam[0].mTargetPos = map(gServoParam[0].start_value, 0, 180, ANGLE_MIN,  ANGLE_MAX);
		break;
	case 1:
		gServoParam[0].mTargetPos = map(gServoParam[0].stop_value, 0, 180, ANGLE_MIN,  ANGLE_MAX);
		break;
	}
	
	//LED, Relay
	pinMode(PIN_RELAY, OUTPUT);
	pinMode(PIN_LED_DIV, OUTPUT);
	pinMode(PIN_LED_STR, OUTPUT);
	
	
	//サーボ出力にセット
	pinMode(PIN_SERVO1, OUTPUT);
	digitalWrite(PIN_SERVO1, HIGH);
	pinMode(PIN_SERVO2, OUTPUT);
	digitalWrite(PIN_SERVO2, HIGH);
	
	myservo.attach(PIN_SERVO1, ANGLE_MIN,  ANGLE_MAX);
	delay(10);
	//myservo.writeMicroseconds(gServoParam[0].mTargetPos);
	delay(10);
	myservo.detach();
	
	//LED, リレー初期設定
	SetLED(gServoParam[0].mStatus);
	SetRelay(gServoParam[0].mStatus);
	SetRelay2(gServoParam[1].mStatus);

	
	
#ifdef DEBUG
	// Serial.println("Init Done");
#endif
	
	gPreviousL6 = 0;
	gPreviousL5 = 0;
}

uint16_t GetIncremental()
{
	uint16_t aInc = analogRead(PIN_SPD) >> 4;
	
	if( aInc < 2)
	{
		aInc = 2;
	}
	else if( aInc > 128)
	{
		aInc = 128;
	}
	
	return aInc;
	
}


void loop()
{
  // You MUST call the NmraDcc.process() method frequently from the Arduino loop() function for correct library operation
  Dcc.process();
  
  if( FactoryDefaultCVIndex && Dcc.isSetCVReady())
  {
    FactoryDefaultCVIndex--; // Decrement first as initially it is the size of the array 
    Dcc.setCV( FactoryDefaultCVs[FactoryDefaultCVIndex].CV, FactoryDefaultCVs[FactoryDefaultCVIndex].Value);
  }
	
	if( (millis() - gPreviousL5) >= 8)
	{
		//Servo control
		Servo_Control(0);
		Servo_Control(1);
		
		//Reset task
		gPreviousL5 = millis();
	}
	
	
	if( (millis() - gPreviousL6) >= 100)
	{
		if( gServoParam[0].mOn == 0)
		{
			
			int aPos2 = map( analogRead(PIN_POS2), 0, 1023, 0, 180 );
			int aPos1 = map( analogRead(PIN_POS1), 0, 1023, 0, 180 );
			int aInc = GetIncremental();
			
			if( (gServoParam[0].stop_value != aPos2) && (gServoParam[0].mOn == 0))
			{
				gServoParam[0].stop_value = aPos2;
				
				if( gServoParam[0].mStatus == 1)
				{
					//サーボ出力にセット
					myservo.attach(PIN_SERVO1, ANGLE_MIN,  ANGLE_MAX);
					gServoParam[0].mTargetPos = map(gServoParam[0].stop_value, 0, 180, ANGLE_MIN,  ANGLE_MAX);
					gServoParam[0].mOn = 1;
				}
			}
			
			if( (gServoParam[0].start_value != aPos1) && (gServoParam[0].mOn == 0))
			{
				gServoParam[0].start_value = aPos1;
				
				if( gServoParam[0].mStatus == 0)
				{
					//サーボ出力にセット
					myservo.attach(PIN_SERVO1, ANGLE_MIN,  ANGLE_MAX);
					gServoParam[0].mTargetPos = map(gServoParam[0].start_value, 0, 180, ANGLE_MIN,  ANGLE_MAX);
					gServoParam[0].mOn = 1;
				}
			}
			
			if( gServoParam[0].increment != aInc)
			{
				gServoParam[0].increment = aInc;
			}
			
		}
	
		//Reset task
		gPreviousL6 = millis();
	}	
	
}

void Servo_MovePos(int inNo, int inPosition)
{
	switch(inNo)
	{
		
	case 0:
		myservo.writeMicroseconds(inPosition);
		break;
		
	case 1:
		myservo2.writeMicroseconds(inPosition);
		break;
	default:
		
		break;

	}
	
}


void Servo_Stop(int inNo)
{
	switch(inNo)
	{
		
	case 0:
		myservo.detach();
		break;
		
	case 1:
		myservo2.detach();
		break;
	default:
		
		break;

	}
	
}

void Servo_Control(int inNo)
{
  	
	if( gServoParam[inNo].mOn == 0)
	{
		return;
	}
	
	if( gServoParam[inNo].current_position >= gServoParam[inNo].mTargetPos)
	{
		//ターゲット位置に到達していない場合
		gServoParam[inNo].current_position = gServoParam[inNo].current_position - gServoParam[inNo].increment;
		
		
		if( gServoParam[inNo].current_position < ANGLE_MIN)
		{
			gServoParam[inNo].current_position = ANGLE_MIN;
		}
		
		if( gServoParam[inNo].current_position <= gServoParam[inNo].mTargetPos)
		{
			//到達判定
			gServoParam[inNo].current_position = gServoParam[inNo].mTargetPos;
			Servo_MovePos(inNo, gServoParam[inNo].mTargetPos);
			//サーボを動かさない
			gServoParam[inNo].mOn = 0;
			
			//CV値に書き込む
			Dcc.setCV( 34 + (inNo * 5), map(gServoParam[inNo].current_position, ANGLE_MIN,  ANGLE_MAX, 0, 180));
			Dcc.setCV( 35 + (inNo * 5), gServoParam[inNo].mStatus);
			
			//サーボ開放
			Servo_Stop(inNo);
		}
		else
		{
			Servo_MovePos(inNo, gServoParam[inNo].current_position);
		}
	}
	else if( gServoParam[inNo].current_position <= gServoParam[inNo].mTargetPos)
	{
		//ターゲット位置に到達していない場合
		gServoParam[inNo].current_position = gServoParam[inNo].current_position + gServoParam[inNo].increment;

		if( gServoParam[inNo].current_position > ANGLE_MAX)
		{
			gServoParam[inNo].current_position = ANGLE_MAX;
		}
		
		
		if( gServoParam[inNo].current_position >= gServoParam[inNo].mTargetPos)
		{
			//到達判定
			gServoParam[inNo].current_position = gServoParam[inNo].mTargetPos;
			Servo_MovePos(inNo, gServoParam[inNo].mTargetPos);
			//サーボを動かさない
			gServoParam[inNo].mOn = 0;
			
			//CV値に書き込む
			Dcc.setCV( 34 + (inNo * 5), map(gServoParam[inNo].current_position, ANGLE_MIN,  ANGLE_MAX, 0, 180));
			Dcc.setCV( 35 + (inNo * 5), gServoParam[inNo].mStatus);
			
			//サーボ開放
			Servo_Stop(inNo);
		}
		else
		{
			Servo_MovePos(inNo, gServoParam[inNo].current_position);
		}
		
	}
	else
	{
		//サーボを動かさない
		gServoParam[inNo].mOn = 0;
	}
	
	
	//// Serial.print("Pos: ");
	//// Serial.println(gServoParam[inNo].current_position);

	
}


void exec_function (int function, int FuncState)
{
	
	
	if( function == 3)
	{
		if( gServoParam[0].mStatus != FuncState)
		{
			
			gServoParam[0].mStatus = FuncState;
			
			//サーボ出力にセット
			myservo.attach(PIN_SERVO1, ANGLE_MIN,  ANGLE_MAX);
			
			switch( FuncState)
			{
			case 0:
				gServoParam[0].mTargetPos = map(gServoParam[0].start_value, 0, 180, ANGLE_MIN,  ANGLE_MAX);
				break;
			case 1:
				gServoParam[0].mTargetPos = map(gServoParam[0].stop_value, 0, 180, ANGLE_MIN,  ANGLE_MAX);
				break;
			}
			
			SetLED(FuncState);
			SetRelay(FuncState);
			
			gServoParam[0].mOn = 1;
		}
		
	}
	
	if( function == 4)
	{	
		if( gServoParam[1].mStatus != FuncState)
		{
			
			gServoParam[1].mStatus = FuncState;
			
			//サーボ出力にセット
			myservo2.attach(PIN_SERVO2, ANGLE_MIN,  ANGLE_MAX);
			
			switch( FuncState)
			{
			case 0:
				gServoParam[1].mTargetPos = map(gServoParam[1].start_value, 0, 180, ANGLE_MIN,  ANGLE_MAX);
				break;
			case 1:
				gServoParam[1].mTargetPos = map(gServoParam[1].stop_value, 0, 180, ANGLE_MIN,  ANGLE_MAX);
				break;
			}
			
			SetRelay2(FuncState);
			gServoParam[1].mOn = 1;

		}
	}
	
}


void SetRelay(uint8_t inState)
{
	
	switch( inState)
	{
	case 0:
		digitalWrite(PIN_RELAY, 0);
		break;
	case 1:
		digitalWrite(PIN_RELAY, 1);
		break;
	default:
		digitalWrite(PIN_RELAY, 0);
		break;
	}	
}

void SetRelay2(uint8_t inState)
{
	
	switch( inState)
	{
	case 0:
		digitalWrite(PIN_RELAY2, 0);
		break;
	case 1:
		digitalWrite(PIN_RELAY2, 1);
		break;
	default:
		digitalWrite(PIN_RELAY2, 0);
		break;
	}	
}

void SetLED(uint8_t inState)
{
	
	switch( inState)
	{
	case 0:
		digitalWrite(PIN_LED_DIV, 0);
		digitalWrite(PIN_LED_STR, 1);
		break;
	case 1:
		digitalWrite(PIN_LED_DIV, 1);
		digitalWrite(PIN_LED_STR, 0);
		break;
	default:
		digitalWrite(PIN_LED_DIV, 0);
		digitalWrite(PIN_LED_STR, 0);
		break;
	}
	
}


uint16_t getMyAddrFromSW()
{
	uint16_t aAddr = 0;
	
	//Address DIP Switch
	pinMode(PI_ADRBIT_0,INPUT);
	pinMode(PI_ADRBIT_1,INPUT);	
	pinMode(PI_ADRBIT_2,INPUT);	
	pinMode(PI_ADRBIT_3,INPUT);	
	pinMode(PI_ADRBIT_4,INPUT);	
	pinMode(PI_ADRBIT_5,INPUT);	
	pinMode(PI_ADRBIT_6,INPUT);	
	pinMode(PI_ADRBIT_7,INPUT);	
	
	digitalWrite(PI_ADRBIT_0, HIGH);
	digitalWrite(PI_ADRBIT_1, HIGH);	
	digitalWrite(PI_ADRBIT_2, HIGH);	
	digitalWrite(PI_ADRBIT_3, HIGH);	
	digitalWrite(PI_ADRBIT_4, HIGH);	
	//digitalWrite(PI_ADRBIT_5, HIGH);	
	digitalWrite(PI_ADRBIT_6, HIGH);	
	digitalWrite(PI_ADRBIT_7, HIGH);	
	
	
	byte aBit0 = NotByte(digitalRead(PI_ADRBIT_0));
	byte aBit1 = NotByte(digitalRead(PI_ADRBIT_1));
	byte aBit2 = NotByte(digitalRead(PI_ADRBIT_2));
	byte aBit3 = NotByte(digitalRead(PI_ADRBIT_3));
	byte aBit4 = NotByte(digitalRead(PI_ADRBIT_4));
	byte aBit5 = NotByte((analogRead(PI_ADRBIT_5) < ANA_ZERO) ? 0 : 1);
	byte aBit6 = NotByte(digitalRead(PI_ADRBIT_6));
	byte aBit7 = NotByte(digitalRead(PI_ADRBIT_7));
	
	
	aAddr += aBit0;
	aAddr += aBit1 << 1;
	aAddr += aBit2 << 2;
	aAddr += aBit3 << 3;
	aAddr += aBit4 << 4;
	aAddr += aBit5 << 5;
	aAddr += aBit6 << 6;
	aAddr += aBit7 << 7;
	
	return aAddr;
	
}

byte NotByte(byte inByte)
{
	if( inByte == 0)
	{
		return 1;
	}
	else
	{
		return 0;
	}
}